import collections.abc
import typing
import re


class Interval(collections.abc.Sequence[str]):
    """
    Expands comma-separated and hyphen-separated string intervals to a list of
    string items with normalized indexation. For example, the interval
    "1A1-1A4, 1A7" is expanded to ["1A01", "1A02", "1A03", "1A04", "1A07"].

    - Each interval item should consist of an alphanumeric "base" followed by
    a numeric "index".
    - A non-empty base can contain any letters and numbers, but must end
    with a letter. Also, the base must be the same across the interval items,
    separated by a hyphen (-).
    - Index must be in the end of the item (thus, it cannot be followed by
    a non-numeric character).

    Args:
        interval (str): An interval to be parsed.

    Raises:
        ValueError: If the input is not a string,
        or if the bases of an interval are not identical.
        AttributeError: If an interval item cannot be split into
        a base and an index.
    """
    INTERVAL_SEPARATORS = ['-', '—']
    INDEX_PATTERN = re.compile("[0-9]+$")

    def __init__(self, interval: typing.Optional[str] = None) -> None:
        super().__init__()
        self._interval = self._check_interval_type(interval)
        self._items = self._parse_to_list()

    def __repr__(self) -> str:
        return self._interval

    def __str__(self) -> str:
        return str(self._items)

    def __getitem__(self, index):
        return self._items[index]

    def __len__(self) -> int:
        return len(self._items)

    @staticmethod
    def _check_interval_type(interval: typing.Optional[str]) -> str:
        if not interval:
            return ''
        if not isinstance(interval, str):
            raise ValueError(
                f'Input interval should be a string, '
                f'got type "{type(interval).__name__}."'
                )
        return interval.strip()

    def _get_sep(self, item: str):
        seps = self.INTERVAL_SEPARATORS
        return next((sep for sep in seps if sep in item), None)

    def _get_item_index(self, item: str) -> int:
        try:
            result = self.INDEX_PATTERN.search(item)
            assert result is not None
            return int(result.group())
        except AssertionError:
            raise TypeError(f'Cannot extract index from item {item}.')

    def _get_item_base(self, item: str):
        try:
            return self.INDEX_PATTERN.sub("", item, count=1)
        except Exception as err:
            raise AttributeError(
                f'Cannot extract interval base from interval {self._interval}.'
                ) from err

    def _parse_to_list(self):
        items = []
        for item in self._interval.split(sep=','):
            item = item.strip()
            if not item:
                continue
            sep = self._get_sep(item)
            if sep is None:
                items.append(item)
                continue
            boundaries = [boundary.strip() for boundary in item.split(sep=sep)]
            boundary_lower = self._get_item_index(boundaries[0])
            boundary_upper = self._get_item_index(boundaries[1])
            interval_base = self._get_item_base(boundaries[0])
            if interval_base != self._get_item_base(boundaries[1]):
                raise ValueError(
                    f'Cannot extract base from interval {self._interval}.'
                    )
            interval_range = range(boundary_lower, boundary_upper + 1)
            for i in interval_range:
                items.append(f'{interval_base}{i:02d}')
        return items



import gspread  # type: ignore
import os
import logging
import pandas
import typing


logger = logging.getLogger('g3tables.utils.gdrive')


class TableMetadataDict(typing.TypedDict):
    id: str
    name: str
    createdTime: str
    modifiedTime: str


@typing.overload
def list_tables_from_gdrive(
    project: str | typing.Any | None, with_meta: typing.Literal[True]
) -> list[TableMetadataDict]:
    ...


@typing.overload
def list_tables_from_gdrive(
    project: str | typing.Any | None, with_meta: typing.Literal[False]
) -> list[str]:
    ...


def list_tables_from_gdrive(
    project: str | typing.Any | None, with_meta: bool = True
) -> list[TableMetadataDict] | list[str]:
    logger.info('Connecting to Google Drive repository')
    creds_file = os.path.join(os.path.dirname(__file__), "creds.json")
    client = gspread.service_account(filename=creds_file)
    logger.info('Collecting Google Drive G3 project tables')
    tables: list[TableMetadataDict] = [
        table for table in client.list_spreadsheet_files()
        ]
    if project is not None:
        logger.info('Filtering "%s" Google Drive G3 project tables', project)
        tables = [
            table for table in tables
            if str(project).casefold() in str(table['name']).casefold()
            ]
    if with_meta:
        return tables
    return [table['name'] for table in tables]


def get_table_data_from_gdrive(
    gdrive_table_name: str,
    header_row: int = 1,
    include_sheets: typing.Optional[typing.Iterable[str]] = None,
    exclude_sheets: typing.Optional[typing.Iterable[str]] = None
) -> dict[str, pandas.DataFrame]:
    """both excluded and included == excluded"""
    logger.info('Connecting to Google Drive repository')
    creds_file = os.path.join(os.path.dirname(__file__), "creds.json")
    client = gspread.service_account(filename=creds_file)
    try:
        logger.info('Loading spreadsheet "%s"', gdrive_table_name)
        table = client.open(gdrive_table_name)
    except gspread.exceptions.SpreadsheetNotFound:
        logger.error(
            'Unable to access spreadsheet "%s". Make sure '
            'that the table name is spelled correctly and '
            'that access to the table has been granted to '
            '"el-script@g3-spreadsheet-sync.iam.gserviceaccount.com".',
            gdrive_table_name
            )
        return {}
    sheets = {}
    for sheet in table.worksheets(exclude_hidden=True):
        if (
            (exclude_sheets and sheet.title in exclude_sheets) or
            (include_sheets and sheet.title not in include_sheets)
        ):
            logger.debug('Ignoring sheet "%s"', sheet.title)
            continue
        try:
            logger.info('Loading sheet "%s"', sheet.title)
            sheets[sheet.title] = pandas.DataFrame(
                sheet.get_all_records(
                    head=header_row, numericise_ignore=['all']
                    )
                )
        except Exception as exp:
            if logger.isEnabledFor(logging.WARNING):  # optimize err str format
                err = f'Failed to load sheet "{sheet.title}"'
                if (exp_str := str(exp)):
                    err = f'{err} ({exp_str})'
                logger.warning(err)
    return sheets



import argparse
import copy
import json
import logging
import os
import re
import typing

from openpyxl import load_workbook, Workbook
from openpyxl.worksheet.worksheet import Worksheet
from openpyxl.cell.cell import Cell
from openpyxl.utils import get_column_letter
from openpyxl.utils import column_index_from_string

from ._table import (
    PLCCompositionIOTable as Table,
    PLCCompositionIOTableColumn as TableColumn
)


logger = logging.getLogger('g3tables.plc_composition_io_table')


def get_input_args():
    parser = argparse.ArgumentParser(
        description=(
            'Prefill the IO table with SW Device placeholders and '
            'the corresponding function block SW connector names.'
            )
        )
    parser.add_argument(
        'table_path',
        type=str,
        help='Path to the HW IO Table'
        )
    parser.add_argument(
        '-re'
        '--add-re-patterns',
        required=False,
        type=str,
        help='Path to the additional regex patterns file'
        )
    return parser.parse_args()


class IOSignalPattern(typing.TypedDict):
    module: str
    connector: str


def get_io_signal_patterns() -> dict[str, IOSignalPattern]:
    file_dir = os.path.dirname(os.path.realpath(__file__))
    file_path = os.path.join(file_dir, 'io_signals.json')
    logger.debug('Loading default IO mapping file "%s"', file_path)
    with open(file_path, 'r') as file:
        return json.load(file)


def get_user_io_signal_patterns(path: str) -> dict[str, IOSignalPattern]:
    logger.debug('Loading user IO mapping file "%s"', path)
    with open(path, 'r') as file:
        patterns = json.load(file)
    logger.debug('Validating user IO mapping patterns data')
    if not isinstance(patterns, dict):
        raise ValueError('Invalid pattern data structure.')
    for signal, data in patterns.items():
        logger.debug('Validating pattern "%s"', signal)
        try:
            module = data['module']
            logger.debug(
                'Retrieved "module=%s" for pattern "%s"', module, signal
                )
            connector = data['connector']
            logger.debug(
                'Retrieved "connector=%s" for pattern "%s"', connector, signal
                )
        except KeyError as err:
            raise ValueError('Invalid pattern data structure.') from err
    return patterns


IO_SIGNAL_PATTERNS = get_io_signal_patterns()


def get_table_name_and_extension(path: str) -> tuple[str, str]:
    logger.debug('Extracting table name and extension from "%s"', path)
    name, ext = os.path.splitext(os.path.basename(path))
    logger.debug('Extracted "name=%s" and "extension=%s"', name, ext)
    return name, ext


def get_table(path: str) -> Workbook:
    logger.debug('Loading "%s" with openpyxl', path)
    return load_workbook(path, data_only=True)


def get_column_index_and_letter(letter_or_index: str | int) -> tuple[int, str]:
    if isinstance(letter_or_index, str):
        logger.debug('Retrieving index for column "%s"', letter_or_index)
        index = column_index_from_string(letter_or_index)
        letter = letter_or_index.upper()
    else:
        logger.debug('Retrieving letter for column "%s"', letter_or_index)
        index = letter_or_index
        letter = get_column_letter(letter_or_index)
    logger.debug('Retrieved "index=%s" and "letter=%s"', index, letter)
    return index, letter


def get_colunm_names(sheet: Worksheet) -> list[str]:
    logger.debug('Extracting table sheet "%s" columns', sheet.title)
    names: list[str] = []
    for i in range(1, sheet.max_column + 1):
        name = sheet.cell(row=Table.HEADER_ROW, column=i).value
        name = str(name) if name else ''
        names.append(name)
    logger.debug('Extracted columns: %s', names)
    return names


def insert_column(
    sheet: Worksheet, index: int | str, name: str, width: int = 20
) -> None:
    logger.debug('Inserting column "%s" at index "%s"', name, index)
    index_int, index_str = get_column_index_and_letter(index)
    index_str_old = get_column_letter(index_int + 1)
    width_old = sheet.column_dimensions[index_str].width
    sheet.insert_cols(index_int)
    logger.debug('Setting column "%s" width to "%s"', index_str, width)
    sheet.column_dimensions[index_str].width = width
    sheet.column_dimensions[index_str_old].width = width_old
    logger.debug('Setting column "%s" name to "%s"', index_str, name)
    header = sheet.cell(row=Table.HEADER_ROW, column=index_int)
    header.value = name


def unmerge_cells_in_new_column(sheet: Worksheet, col_idx: int):
    """
    Must be placed after the formatting functions. For some reason,
    crashes if used before the formatting of both columns (KeyError).
    """
    affected_ranges = []
    for merged_range in sheet.merged_cells.ranges:
        min_col, min_row, max_col, max_row = merged_range.bounds
        if min_col <= col_idx <= max_col and max_row > Table.HEADER_ROW:
            affected_ranges.append(merged_range)
    for merged_range in affected_ranges:
        sheet.unmerge_cells(str(merged_range))


def format_column(
    sheet: Worksheet, column_index_new: int, column_index_reference: int
) -> None:
    logger.debug(
        'Copying column cells\' formatting from "%s" to "%s"',
        get_column_letter(column_index_reference),
        get_column_letter(column_index_new)
        )
    for row in range(1, sheet.max_row + 1):
        cell_ref = sheet.cell(row=row, column=column_index_reference)
        cell_new = sheet.cell(row=row, column=column_index_new)
        if cell_ref.has_style:
            cell_new.font = copy.copy(cell_ref.font)
            cell_new.border = copy.copy(cell_ref.border)
            cell_new.number_format = copy.copy(cell_ref.number_format)
            cell_new.protection = copy.copy(cell_ref.protection)
            cell_new.alignment = copy.copy(cell_ref.alignment)
            column = Table.get_column_index(TableColumn.TERMINAL)
            if not sheet.cell(row=row, column=column).value:
                cell_new.fill = copy.copy(cell_ref.fill)
                continue


def extend_sheet(
    sheet: Worksheet,
    user_patterns: typing.Optional[dict[str, IOSignalPattern]] = None
) -> None:

    not_matched_symbol = "?"

    def is_cell_empty(cell: Cell | typing.Any) -> bool:
        return not cell.value or cell.value == not_matched_symbol

    def match_pattern(
        sw_device_type: Cell | typing.Any,
        sw_device_name: Cell | typing.Any,
        sw_signal: Cell | typing.Any,
        sw_signal_alias: Cell | typing.Any,
        patterns: dict[str, IOSignalPattern]
    ) -> bool:
        for pattern, data in patterns.items():
            matched = bool(re.search(pattern, str(sw_signal_alias.value)))
            if not matched:
                continue
            if (is_cell_empty(sw_device_type) and is_cell_empty(sw_signal)):
                module_name_full = data['module']
                module_name_last = module_name_full.split('/')[-1]
                sw_device_type.value = module_name_full
                sw_device_name.value = f"<{module_name_last} name>"
                sw_signal.value = data['connector']
            break
        return matched

    for row in range(Table.HEADER_ROW + 1, sheet.max_row + 1):
        io_name_cell = sheet.cell(
            row=row, column=Table.get_column_index(TableColumn.SIGNAL_NAME)
            )
        sw_device_type_cell = sheet.cell(
            row=row, column=Table.get_column_index(TableColumn.SW_DEVICE_TYPE)
            )
        sw_device_name_cell = sheet.cell(
            row=row, column=Table.get_column_index(TableColumn.SW_DEVICE_NAME)
            )
        sw_signal_cell = sheet.cell(
            row=row, column=Table.get_column_index(TableColumn.SW_SIGNAL)
            )
        signal_cell = sheet.cell(
            row=row, column=Table.get_column_index(TableColumn.SIGNAL)
            )
        if (
            not io_name_cell.value or
            not signal_cell.value or
            'safe' in str(io_name_cell.value).lower() or
            'reserve' in str(signal_cell.value).lower()
        ):
            continue
        matched = False
        if user_patterns:
            matched = match_pattern(
                sw_device_type_cell,
                sw_device_name_cell,
                sw_signal_cell,
                signal_cell,
                user_patterns
                )
        if not matched:
            matched = match_pattern(
                sw_device_type_cell,
                sw_device_name_cell,
                sw_signal_cell,
                signal_cell,
                IO_SIGNAL_PATTERNS
                )
        if not matched:
            if (
                is_cell_empty(sw_device_type_cell) and
                is_cell_empty(sw_signal_cell)
            ):
                sw_device_type_cell.value = not_matched_symbol
                sw_device_name_cell.value = f'<{not_matched_symbol}>'
                sw_signal_cell.value = not_matched_symbol


def extend_workbook(
    table: Workbook,
    user_patterns: typing.Optional[dict[str, IOSignalPattern]] = None,
) -> None:
    for sheet in table.worksheets:
        if sheet.title in Table.EXCLUDE_SHEETS:
            logger.debug('Ignoring table sheet "%s"', sheet.title)
            continue
        logger.debug('Extending table sheet "%s"', sheet.title)
        column_names = get_colunm_names(sheet)
        logger.debug(
            'Checking if column "%s" is present in the table', 'SWDeviceType'
            )
        # temp insert: the 'Cabinet' column is inserted by
        # PLCCompositionIOTable during table data parsing at runtime
        insert_column(
            sheet,
            Table.get_column_index(TableColumn.CABINET),
            'Temp'
            )
        if 'SWDeviceType' not in column_names:
            insert_column(
                sheet,
                Table.get_column_index(TableColumn.SW_DEVICE_TYPE),
                'SWDeviceType'
                )
            format_column(
                sheet,
                Table.get_column_index(TableColumn.SW_DEVICE_TYPE),
                Table.get_column_index(TableColumn.SW_DEVICE_TYPE) - 1,
                )
        logger.debug(
            'Checking if column "%s" is present in the table', 'SWDeviceName'
            )
        if 'SWDeviceName' not in column_names:
            insert_column(
                sheet,
                Table.get_column_index(TableColumn.SW_DEVICE_NAME),
                'SWDeviceName'
                )
            format_column(
                sheet,
                Table.get_column_index(TableColumn.SW_DEVICE_NAME),
                Table.get_column_index(TableColumn.SW_DEVICE_NAME) - 1,
                )
        logger.debug(
            'Checking if column "%s" is present in the table', 'SWSignal'
            )
        if 'SWSignal' not in column_names:
            insert_column(
                sheet,
                Table.get_column_index(TableColumn.SW_SIGNAL),
                'SWSignal'
                )
            format_column(
                sheet,
                Table.get_column_index(TableColumn.SW_SIGNAL),
                Table.get_column_index(TableColumn.SW_SIGNAL) - 1,
                )
        signal_col_letter = get_column_letter(
            Table.get_column_index(TableColumn.SIGNAL) - 1
            )  # -1 because the temp column before this column is deleted later
        remark_col_letter = get_column_letter(
            Table.get_column_index(TableColumn.REMARK) - 1
            )  # -1 because the temp column before this column is deleted later
        sheet.column_dimensions[signal_col_letter].width = 20
        sheet.column_dimensions[remark_col_letter].width = 60
        # NOTE: cannot place the unmerging before the formatting
        unmerge_cells_in_new_column(
            sheet, Table.get_column_index(TableColumn.SW_DEVICE_NAME)
            )
        unmerge_cells_in_new_column(
            sheet, Table.get_column_index(TableColumn.SW_DEVICE_TYPE)
            )
        unmerge_cells_in_new_column(
            sheet, Table.get_column_index(TableColumn.SW_SIGNAL)
            )
        extend_sheet(sheet, user_patterns)
        # remove the temp insert
        logger.debug(
            'Deleting column "%s" at index "%s"',
            'Temp', Table.get_column_index(TableColumn.CABINET)
            )
        sheet.delete_cols(1)


def extend_table(
    file_path: str,
    user_patterns_file_path: str | None = None,
    output_dir_path: str | None = None,
    output_file_name: str | None = None
) -> None:
    # load source table
    logger.info('Loading PLC Composition and IO Table "%s"', file_path)
    table = get_table(file_path)
    # load user regex patterns
    user_patterns: dict[str, IOSignalPattern] = {}
    if user_patterns_file_path:
        logger.info(
            'Extracting user IO mapping patterns from "%s"',
            user_patterns_file_path
            )
        user_patterns = get_user_io_signal_patterns(user_patterns_file_path)
    # extend the table columns
    logger.info('Extending PLC Composition and IO Table "%s"', file_path)
    extend_workbook(table, user_patterns)
    # save the updated table
    output_dir_path = output_dir_path or os.getcwd()
    if not output_file_name:
        table_name, table_ext = get_table_name_and_extension(file_path)
        output_file_name = f'{table_name} Extended{table_ext}'
    output_file_path = os.path.join(output_dir_path, output_file_name)
    logger.info(
        'Saving the extended PLC Composition and IO Table to "%s"',
        output_file_path
        )
    table.save(output_file_path)


def main() -> None:
    args = get_input_args()
    extend_table(args.table_path, args.add_re_patterns)


if __name__ == "__main__":
    main()



import pandas as pd
import logging
import typing
import warnings

from ..utils.gdrive import get_table_data_from_gdrive


logger = logging.getLogger('g3tables.visualization_table')


DEVICE_TYPES = {
    "System": "System_G3",
    "SystemSafety": "SystemSafety_G3",
    "Zone": "Zone_G3",
    "Gate": "Gate_G3",
    "Route": "Route_G3",
    "Zone.Track": "Track",
    "Detector.TrackCircuit": "TC_G3",
    "Detector.Pantograph": "PD_G3",
    "PointMachine.PME": "PME_G3",
    "PointMachine.PMM": "PMM_G3",
    "Signal": "Signal_G3",
    "Signal.Symbol": "SignalSymbol_G3",
    "DoorDisplay7": "DoorDisplay7_G3",
    "Requestor.Digital": "RequestorDigital",
    "Requestor.RoutingTable": "RoutingTable_G3",
    "Requestor.Vecom": "VecomController_G3",
    "Requestor.Vecom.Loop": "VecomLoop_G3",
    "Requestor.DRR": "DRRController_G3",
    "Requestor.DRR.Transceiver": "DRRTransceiver_G3",
    "Requestor.SPIE": "SPIEController_G3",
    "Requestor.SPIE.Loop": "SPIELoop_G3",
    "Requestor.Vetra": "Vetra_G3",
    "Requestor.Digital.AWA": "AWA_G3",
    "Cabinet": "Cabinet_G3",
    "Cabinet.UPS": "CabinetUps_G3",
    "Cabinet.Fuse": "CabinetFuse_G3",
    "Cabinet.Convertor": "CabinetConvertor_G3",
    "Cabinet.MonitoringModule": "",
    "Heating": "Heating_G3",
    "Heating.Contactor": "HeatingContactor_G3",
    "Heating.Contactor.Rod": "HeatingRod_G3",
    "Matrix": "MPS_G3",
    "GPIO": "GPIO_G3"
}


class SHVVarDict(typing.TypedDict):
    name: str
    is_blacklisted: bool
    is_deleted: bool


class SHVProjectTypeDict(typing.TypedDict):
    type: str
    restricted_to: str | None
    vars: dict[str, SHVVarDict]


class VisualizationTable:
    NAME_PATTERN = 'project_visualization'
    EXCLUDED_SHEETS = [
        'Title', 'Identification', 'Worksheet explanation', 'Users'
        ]

    def __init__(self, params_sheet_data: pd.DataFrame) -> None:
        self._data = params_sheet_data

    @classmethod
    def is_name_valid(cls, name: str) -> bool:
        name_formatted = name.strip().replace(' ', '_').casefold()
        return cls.NAME_PATTERN in name_formatted

    @classmethod
    def from_local(cls, path: str) -> typing.Self:
        with warnings.catch_warnings():
            warnings.filterwarnings('ignore', category=UserWarning)
            params_sheet = pd.read_excel(
                path, header=2, sheet_name='Parameters'
                )
        formatter = VisualizationTableFormatter.format_params_sheet
        return cls(formatter(params_sheet))

    @classmethod
    def from_gdrive(cls, gdrive_table_name: str) -> typing.Self:
        sheets_all = get_table_data_from_gdrive(
            gdrive_table_name, header_row=3, exclude_sheets=cls.EXCLUDED_SHEETS
            )
        formatter = VisualizationTableFormatter.format_params_sheet
        return cls(formatter(sheets_all['Parameters']))

    @classmethod
    def load(cls, path: str) -> typing.Self:
        if path.endswith('.xlsx'):
            return cls.from_local(path)
        return cls.from_gdrive(path)

    @property
    def data(self) -> pd.DataFrame:
        return self._data

    @staticmethod
    def _new_shv_project_type_dict(
        module_type: str, restricted_to: str | None
    ) -> SHVProjectTypeDict:
        return {
            'type': module_type,
            'restricted_to': restricted_to,
            'vars': {}
            }

    @staticmethod
    def _new_shv_var_dict(
        shv_var_name: str, is_blacklisted: bool, is_deleted: bool
    ) -> SHVVarDict:
        return {
            'name': shv_var_name,
            'is_blacklisted': is_blacklisted,
            'is_deleted': is_deleted
            }

    @property
    def project_types(self) -> dict[str, SHVProjectTypeDict]:
        project_types: dict[str, SHVProjectTypeDict] = {}
        for _, row in self._data.iterrows():
            # create or fetch existing project_type_data dictionary
            project_type = row['Project type']
            project_subtype = row['Project subtype']
            if project_subtype:
                module_type, restricted_to = project_subtype, project_type
            else:
                module_type, restricted_to = project_type, None
            project_type_data = project_types.setdefault(
                module_type,
                self._new_shv_project_type_dict(module_type, restricted_to)
                )
            # create shv variable data dictionary
            if not (shv_var := row['SHV variable']):
                continue
            project_type_data['vars'][shv_var] = self._new_shv_var_dict(
                shv_var_name=shv_var,
                is_blacklisted=row['Available only in SHVspy'],
                is_deleted=row['x']
                )
        return project_types


class VisualizationTableFormatter:

    @staticmethod
    def fill_in_device_types(device):
        if device in DEVICE_TYPES:
            return DEVICE_TYPES[device]
        logging.warning(
            f'Could not retrieve SHV Device type for module "{device}".'
            )
        return ''

    @staticmethod
    def format_params_sheet(params_sheet: pd.DataFrame) -> pd.DataFrame:
        # drop rows where the 'Module (device)' column value is empty
        params_sheet = params_sheet.fillna('').astype(str, copy=True)
        params_sheet_mask = (params_sheet['Module (device)'] != '')
        params_sheet = params_sheet[params_sheet_mask]
        # create the 'Project type' column
        params_sheet['Project type'] = params_sheet['Module (device)'].apply(
            VisualizationTableFormatter.fill_in_device_types
            )
        # format the 'x' column to bool
        params_sheet.loc[:, 'x'] = params_sheet['x'].apply(
            lambda cell_val: bool(cell_val)
        )
        # format the 'Available only in SHVspy' column to bool
        col_name = 'Available only in SHVspy'
        params_sheet.loc[:, col_name] = params_sheet[col_name].apply(
            lambda cell_val: cell_val == 'TRUE'
        )
        return params_sheet



from ._table import (
    PLCCompositionIOTable,
    PLCCompositionIOTableColumn,
    PLCCompositionIOTableColumnIndex
)
from ._extend_table import extend_table, IO_SIGNAL_PATTERNS


__all__ = [
    'PLCCompositionIOTable',
    'PLCCompositionIOTableColumn',
    'PLCCompositionIOTableColumnIndex',
    'extend_table',
    'IO_SIGNAL_PATTERNS'
]



import typing


def to_nested_dict(
    data: typing.Mapping[str, typing.Mapping | typing.Any], keysep: str
) -> dict[str, dict | typing.Any]:
    """
    Converts a dictionary with composite keys separated by
    a specified separator into a nested dictionary.
    The function handles nested dictionaries recursively.

    Example:
    ```
    init_dict = {
        'a/b/c': 1,
        'a/b/d': {'e': 1},
        'a/b/f': {'g/h': 1}
    }
    to_nested_dict(init_dict, sep='/')
    >>> {'a': {'b': {'c': 1, 'd': {'e': 1}, 'f': {'g': {'h': 1}}}}}
    ```

    Args:
        data (typing.Mapping[str, typing.Mapping | typing.Any]):
        Dictionary to be converted. Keys are strings with a separator.
        sep (str): Separator used in the keys of the dictionary.

    Returns:
        dict[str, dict | typing.Any]: Nested dictionary where
        the structure is defined by the composite keys.
    """

    def recursive_convert(data: dict | typing.Any) -> dict | typing.Any:
        if not isinstance(data, dict):
            return data
        return {
            key: recursive_convert(value)
            for key, value in to_nested_dict(data, keysep).items()
            }

    nested_dict: dict[str, dict | typing.Any] = {}
    for key, value in data.items():
        keys = key.split(keysep)
        sub_dict = nested_dict
        for sub_key in keys[:-1]:
            sub_dict = sub_dict.setdefault(sub_key, {})
        sub_dict[keys[-1]] = recursive_convert(value)
    return nested_dict
